/*
 * Copyright (c) 2001,2002,2003
 *    Digital Stratum.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Company, Digital Stratum, nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Moves records from the active log table into the temp table which
 * has indexes.  Also facilitates creation of Combined Log Format log
 * files from the temp table.
 *
 * TABLE FORMAT
 *
 * Active (no indexes)
 *
                   Table "log_entries"
      Column      |           Type           | Modifiers
------------------+--------------------------+-----------
 request_time     | timestamp with time zone |
 seconds_to_serve | integer                  |
 response_code    | integer                  |
 bytes_sent       | integer                  |
 request_type     | character varying(10)    |
 virtual_domain   | character varying(80)    |
 remote_host      | character varying(80)    |
 request_url_path | character varying(80)    |
 referer          | text                     |
 user_agent       | text                     |
 request          | text                     |
 logname          | character varying(30)    |
 username         | character varying(30)    |
 *
 *
 * Stats (full info plus indexes)
 *
                  Table "temp_entries"
      Column      |           Type           | Modifiers
------------------+--------------------------+-----------
 request_time     | timestamp with time zone |
 seconds_to_serve | integer                  |
 response_code    | integer                  |
 bytes_sent       | integer                  |
 request_type     | character varying(10)    |
 virtual_domain   | character varying(80)    |
 remote_host      | character varying(80)    |
 request_url_path | character varying(80)    |
 referer          | text                     |
 user_agent       | text                     |
 request          | text                     |
 logname          | character varying(30)    |
 username         | character varying(30)    |
Indexes: temp_entries_remotehost,
         temp_entries_requesttime,
         temp_entries_responsecode,
         temp_entries_virtualdomain
 *
 * gcc -o pglogctl -I/usr/local/pgsql/include -L/usr/local/pgsql/lib -lpq pglogctl.c
 */

/* >>> BEGIN USER CONFIGURATION <<< */

#define DBNAME "pglogd"
#define CONNECTION "host=localhost dbname=pglogd user=postgres"
/* #define CONNECTION "hostaddr=10.0.0.1 port=5432 dbname=pglog user=postgres" */

/* >>> END USER CONFIGURATION <<< */


#include <libpq-fe.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>


#define DOMAINSIZE	80		/* Domain names over 80 characters are going to be chopped. */
#define QUERYSIZE	8192


/*
 * Prototype functions
 */
void checkConnStatus(PGconn *pgConn, const char *dbName, char *cpProg);
void checkResStatus(PGconn *pgConn, PGresult *pgRes, char *cpProg, char *cpQuery);
void show_usage(void);

/*
 * Main
 */
int
main(int argc, char **argv)
{
	PGconn		*pgConn;		/* Database Connection */
	PGresult	*pgRes;			/* Handle to a query result set */

	char cStart[30];			/* Start timestamp. */
	char cStop[30];				/* Stop timestamp. */
	char cQuery[QUERYSIZE];		/* Query buffer. */
	char cDomain[DOMAINSIZE];	/* Maximun domain name length. */

	char *cpValue;

	int iAction;
	int iDays;
	int iTZ;
	int iTZminute;
	int i;


	// Make sure we have the correct number parameters.
	if ( argc != 5 )
	{
		show_usage();
		exit(1);
	}

	// Determine what we will be doing.
	iAction = 0;

	if ( strncmp("-o", argv[1], 2) == 0 )
		iAction = 1;
	else if ( strncmp("-m", argv[1], 2) == 0 )
		iAction = 2;
	else if ( strncmp("-d", argv[1], 2) == 0 )
		iAction = 3;
	else if ( strncmp("-p", argv[1], 2) == 0 )
		iAction = 4;

	// Make sure we have *something* for a domain name.
	// A minimum domain would be: x.yy
	if ( strlen(argv[2]) <= 3 )
	{
		fprintf(stderr, "%s: <%s> is too short.\n", argv[0], argv[2]);
		exit(2);
	}

	// Make sure the domain name is not too long.
	strncpy(cDomain, argv[2], DOMAINSIZE);
	cDomain[DOMAINSIZE] = '\0';

	// Check the date length.
	if ( strlen(argv[3]) != 10 )
	{
		fprintf(stderr, "%s: Start Date error.  Format must be mm/dd/yyyy or mm.dd.yyyy\n", argv[0]);
		exit(3);
	}

	// Make an integer out of the days parameter.
	iDays = atoi(argv[4]);

	if ( iDays > 0 )
		iDays--;

	// Connect to the database.
	pgConn = PQconnectdb(CONNECTION);
	checkConnStatus(pgConn, DBNAME, argv[0]);


	// Make a full start and stop date.  Instead of messing with
	// a bunch of date math and libs, we will just use the
	// database to give us a proper timestamp.
	if ( iDays >= 0 )
		snprintf(cQuery, QUERYSIZE, "select '%s'::timestamp as date", argv[3]);
	else
		snprintf(cQuery, QUERYSIZE, "select ('%s'::timestamp + '%d days') as date", argv[3], iDays);

	pgRes = PQexec(pgConn, cQuery);
	checkResStatus(pgConn, pgRes, argv[0], cQuery);
	cpValue = PQgetvalue(pgRes, 0, 0);
	strncpy(cStart, cpValue, 30);
	PQclear(pgRes);

	if ( iDays >= 0 )
		snprintf(cQuery, QUERYSIZE, "select ('%s'::timestamp + '%d days') as date", cStart, iDays);
	else
		snprintf(cQuery, QUERYSIZE, "select ('%s'::timestamp + '-1 days') as date", argv[3]);

	pgRes = PQexec(pgConn, cQuery);
	checkResStatus(pgConn, pgRes, argv[0], cQuery);
	cpValue = PQgetvalue(pgRes, 0, 0);
	strncpy(cStop, cpValue, 30);
	PQclear(pgRes);

	// Adjust the stop timestamp to 23:59:59.  This is the
	// assumed format: Mon Jan 01 00:00:00 2001 EST
	bcopy("23:59:59", &cStop[11], 8);

	// Check the action.
	if ( iAction == 1 )
	{
		// Run the query and dump the results.
		snprintf(cQuery, QUERYSIZE, "SELECT remote_host,logname,username,"
			"to_char(request_time, 'DD/Mon/YYYY:HH24:MI:SS'),"
			"extract(timezone_hour from request_time),"
			"extract(timezone_minute from request_time),request,response_code,bytes_sent,referer,user_agent "
			"FROM temp_entries WHERE virtual_domain='%s' "
			"AND request_time BETWEEN '%s' AND '%s' ORDER BY request_time",	cDomain, cStart, cStop);
		pgRes = PQexec(pgConn, cQuery);
		checkResStatus(pgConn, pgRes, argv[0], cQuery);

		// Output extended CLF
		// "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-agent}i\""
		for ( i = 0 ; i < PQntuples(pgRes) ; i++ )
		{
			iTZ = (atoi(PQgetvalue(pgRes, i, 4)) * -1);
			iTZminute = atoi(PQgetvalue(pgRes, i, 5));

			fprintf(stdout, "%s %s %s [%s %0+3d%02d] \"%s\" %s %s \"%s\" \"%s\"\n",
				PQgetvalue(pgRes, i, 0), // remote host ip
				PQgetvalue(pgRes, i, 1), // log name
				PQgetvalue(pgRes, i, 2), // username
				PQgetvalue(pgRes, i, 3), // time stamp
				iTZ, iTZminute,          // Timezone
				PQgetvalue(pgRes, i, 6), // request
				PQgetvalue(pgRes, i, 7), // status
				PQgetvalue(pgRes, i, 8), // bytes transfered
				PQgetvalue(pgRes, i, 9), // referer
				PQgetvalue(pgRes, i, 10) // user agent
				);
		}

		PQclear(pgRes);
	}

	else if ( iAction == 2 )
	{
		// Move to temp entries table.
		snprintf(cQuery, QUERYSIZE, "INSERT INTO temp_entries SELECT * FROM log_entries "
				"WHERE virtual_domain='%s' AND request_time BETWEEN '%s' AND '%s'", cDomain, cStart, cStop);
		pgRes = PQexec(pgConn, cQuery);
		checkResStatus(pgConn, pgRes, argv[0], cQuery);
		fprintf(stdout, "Log entries moved to temp entries: [%s]\n", PQcmdTuples(pgRes));

		PQclear(pgRes);

		// Delete from active table.
		snprintf(cQuery, QUERYSIZE, "DELETE FROM log_entries WHERE virtual_domain='%s' "
			"AND request_time BETWEEN '%s' AND '%s'", cDomain, cStart, cStop);
		pgRes = PQexec(pgConn, cQuery);
		checkResStatus(pgConn, pgRes, argv[0], cQuery);
		fprintf(stdout, "Log entries removed (should match number moved): [%s]\n", PQcmdTuples(pgRes));

		PQclear(pgRes);

	}

	else if ( iAction == 3 )
	{
		// Delete.
		snprintf(cQuery, QUERYSIZE, "DELETE FROM temp_entries WHERE virtual_domain='%s' "
			"AND request_time BETWEEN '%s' AND '%s'", cDomain, cStart, cStop);
		pgRes = PQexec(pgConn, cQuery);
		checkResStatus(pgConn, pgRes, argv[0], cQuery);
		fprintf(stdout, "Temp entries removed: [%s]\n", PQcmdTuples(pgRes));

		PQclear(pgRes);
	}

	else if ( iAction == 4 )
	{
		// Print the values we come up with.
		snprintf(cQuery, QUERYSIZE, "SELECT count(*) FROM temp_entries "
				"WHERE virtual_domain='%s' AND request_time BETWEEN '%s' AND '%s'", cDomain, cStart, cStop);
		pgRes = PQexec(pgConn, cQuery);
		checkResStatus(pgConn, pgRes, argv[0], cQuery);
		fprintf(stdout, "Temp entries found: [%s]\n", PQgetvalue(pgRes, 0, 0));

		PQclear(pgRes);

		// Adjust iDays for proper display.
		if ( iDays < 0 )
			iDays = abs(iDays);
		else
			iDays++;

		printf("Domain: [%s]\nStart Date: [%s]\nStop Date: [%s]\nDays: [%d]\n", cDomain, cStart, cStop, iDays);
	}

	else
	{
		// Undefined action.
		show_usage();
		exit(4);
	}


	// Close the database connection.
	PQfinish(pgConn);

	return(0);
}
// main()


/*
 * Checks the status of a database connect.  Exits if there was an error
 */
void
checkConnStatus(PGconn *pgConn, const char *dbName, char *cpProg)
{
	if ( PQstatus(pgConn) == CONNECTION_BAD )
	{
		fprintf(stderr, "%s: Connection to database <%s> failed.\n", cpProg, dbName);
		fprintf(stderr, "%s: %s", cpProg, PQerrorMessage(pgConn));
		PQfinish(pgConn);
		exit(4);
	}

	return;
}
// checkConnStatus()


/*
 * Checks the status of a query.  Exits if there was an error
 */
void
checkResStatus(PGconn *pgConn, PGresult *pgRes, char *cpProg, char *cpQuery)
{
	if ( !pgRes || ((PQresultStatus(pgRes) != PGRES_EMPTY_QUERY) &&
					(PQresultStatus(pgRes) != PGRES_COMMAND_OK) &&
					(PQresultStatus(pgRes) != PGRES_TUPLES_OK)) )
	{
		fprintf(stderr, "%s: Query Failed\n", cpProg);
		fprintf(stderr, "%s: Offending Query - <%s>\n", cpProg, cpQuery);
		fprintf(stderr, "%s: %s", cpProg, PQresultErrorMessage(pgRes));
		PQclear(pgRes);
		PQfinish(pgConn);
		exit(5);
	}

	return;
}
// checkResStatus()


/*
 * Display Usage.
 */
void
show_usage(void)
{
	fprintf(stderr,
		"\nMoves records from the log entries table into the temp entries table which\n"
		"has indexes.  Also facilitates creation of Combined Log Format \n"
		"log files from the temp table.\n\n"
 		"Useage:\n\n"
 		"pglogctl [-o | -m | -d | -p] domain startdate days\n\n"
 		"\t-o\tOutput Combined Log Format from the temp entries table\n"
 		"\t-m\tMove records from log entries table to the temp entries table\n"
 		"\t-d\tDelete records from the temp entries table\n"
 		"\t-p\tPrint values to be used based on command line options.\n\n"
 		"\tdomain\t\tDomain to use, required.\n"
 		"\tstartdate\tFormat: mm/dd/yyyy or mm.dd.yyyy\n"
 		"\t\t\tStart time will be 00:00:00 and end time will be 23:59:59\n"
 		"\tdays\t\tNumber of days, including the start date, to process.\n"
 		"\t\t\tValue can be negative.\n\n");
	return;
}
// show_usage()
