/*
 * Copyright (c) 2001, 2002, 2003
 *    Digital Stratum.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Company, Digital Stratum, nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * PGLOGD v2.3
 *
 * Log web server entries to a PostgreSQL database.
 */

#ifndef _PGLOGD_H_
#define _PGLOGD_H_

#define QUERYBASE	512		/* Query size not including variable length data. */
#define BUFFSIZE	16384	/* This should be reasonable to hold a log entry. */
#define MAXFILES	255		/* Maximum files open by a uid, if not specified. */

#define TRUE 1
#define FALSE 0

#define CR '\r'
#define LF '\n'
#define CRLF '\r\n'

#define ERR_NORMAL			0		/* Normal exit. */
#define ERR_FIFOSTAT		-1		/* Error when checking FIFO with stat(). */
#define ERR_FIFOEXISTS		-2		/* The FIFO already exists. */
#define ERR_FIFOCREATE		-3		/* Could not create FIFO. */
#define ERR_FIFOOPEN		-4		/* Could not open FIFO. */
#define ERR_FIFOREAD		-5		/* Could not read the FIFO. */
#define ERR_OVERFLOW		-6		/* Could not open or create the overflow file. */
#define ERR_OVERFLOWREAD	-7		/* Could not read from or find and entry in the overflow file. */
#define ERR_OVERFLOWWRITE	-8		/* Could not write to the overflow file. */
#define ERR_DBCONN			-9		/* Database connection failure. */
#define ERR_DBQUERY			-10		/* Database query failure. */
#define ERR_DBNBLOCK		-11		/* Failure setting the database connection to non-blocking. */
#define ERR_FORK			-12		/* Could not fork(). */
#define ERR_CATCH_SIGHUP	-13		/* Error trying to catch SIGHUP. */
#define ERR_CATCH_SIGINT	-14		/* Error trying to catch SIGINT. */
#define ERR_CATCH_SIGTERM	-15		/* Error trying to catch SIGTERM. */
#define ERR_SELECT			-16		/* Error calling select(). */
#define ERR_MEMORY			-17		/* System memory error, usually associated with malloc(). */


/* User Configuration Defaults */
#define DBNAME "pglogd"
#define CONNECTION "dbname=pglogd user=postgres" /* Local database, trusted. */
//#define CONNECTION "hostaddr=10.0.0.1 port=5432 dbname=pglogd user=postgres" /* Remote database. */

#define FIFO "/tmp/.f.PGLOGD"			/* FIFO file. */
#define LOGFILE "/var/log/pglogd.log"	/* File the daemon will log to. */

#define OVERFLOWFILE "/usr/site/logs/pglogd_overflow"
#define OVERFLOW_WARNING 100			/* Write a warning to the log every n entries in the overflow log. */

#define MAXSELECTDELAY 30				/* Max time select() will block. */
#define TLDONLY TRUE					/* True to strip a virtual back to the top level domain only. */



/*
 * Structure for user config file settings.  Only one of these
 * will exist and it will be global.
 */
struct user_config {
	char cDBNAME[256];
	char cCONNECTION[256];
	char cFIFO[256];
	char cLOGFILE[256];
	char cOVERFLOWFILE[256];
	int iOVERFLOW_WARNING;
	int iMAXSELECTDELAY;
	int iTLDONLY;
};


static char cKeywords[][24] = {
	"database_name",
	"connect_string",
	"fifo_file",
	"log_file",
	"overflow_file",
	"overflow_warning_level",
	"max_select_delay",
	"top_level_domain_only"
};

#define NUM_KEYWORDS (sizeof(cKeywords) / 24)


enum eKeywords {
	dbname,
	connection,
	fifo,
	logfile,
	overflowfile,
	overflow_warning,
	maxselectdelay,
	tldonly
};


/*
 * Parsed pglogd format log entry.
 *
 * original, referer, user_agent, and request are dynamically
 * allocated and must be free()'d with free_entry() only!
 *
 * Character fields are 1 greater than the equivalent
 * fields in the datatable.  This is because the string
 * terminator is accounted for in the database, but
 * we must deal with it ourselves.
 */
struct pglog_entry {
	char *original;
	char request_time[31];
	int seconds_to_serve;
	int response_code;
	int bytes_sent;
	char request_type[11];
	char virtual_domain[81];
	char remote_host[81];
	char *request_url_path;
	char *referer;
	char *user_agent;
	char *request;
	char logname[31];
	char username[31];
};


/*
 * Prototype functions.
 */
int checkConnStatus(PGconn *pgConn, const char *dbName, char *cpProg);
int checkResStatus(PGconn *pgConn, PGresult *pgRes, char *cpProg, char *cpQuery);
int escapeSpecial(char *cpIn, char *cpOut);
int open_overflow(int *iRead, int *iWrite, int iMode);
int init_overflow(int iRead);
int parse_entry(char *cpReadBuffer, struct pglog_entry *spLogEntry);
void free_entry(struct pglog_entry *spLogEntry);
static void sig_init(int iSig);
void logentry(const char *cpError, int iErrno);
void read_config(const char *cpFile);
int strtolower(char *cpString);
int true_or_false(const char *cpString);

#endif /* !_PGLOGD_H_ */
